@@ -31,8 +31,6 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #define NULL 0
 
-static tree build_up_reference ();
-
 /* Change of width--truncation and extension of integers or reals--
    is represented with NOP_EXPR.  Proper functioning of many things
    assumes that no other conversions can be NOP_EXPRs.
@@ -208,13 +206,14 @@ build_up_reference (type, arg, flags)
      tree type, arg;
      int flags;
 {
-  tree rval;
+  tree rval, targ;
   int literal_flag = 0;
   tree argtype = TREE_TYPE (arg), basetype = argtype;
   tree target_type = TREE_TYPE (type);
 
   assert (TREE_CODE (type) == REFERENCE_TYPE);
-  if (TYPE_MAIN_VARIANT (argtype) != TYPE_MAIN_VARIANT (target_type)
+  if (flags != 0
+      && TYPE_MAIN_VARIANT (argtype) != TYPE_MAIN_VARIANT (target_type)
       && IS_AGGR_TYPE (argtype)
       && IS_AGGR_TYPE (target_type))
     {
@@ -229,20 +228,24 @@ build_up_reference (type, arg, flags)
 	}
     }
 
-  switch (TREE_CODE (arg))
+  targ = arg;
+  if (TREE_CODE (targ) == SAVE_EXPR)
+    targ = TREE_OPERAND (targ, 0);
+
+  switch (TREE_CODE (targ))
     {
     case INDIRECT_REF:
       /* This is a call to a constructor which did not know what it was
 	 initializing until now: it needs to initialize a temporary.  */
       if (TYPE_HAS_CONSTRUCTOR (arg))
 	{
-	  tree temp = build_cplus_new (argtype, TREE_OPERAND (arg, 0));
-	  TYPE_HAS_CONSTRUCTOR (arg) = 0;
+	  tree temp = build_cplus_new (argtype, TREE_OPERAND (targ, 0), 1);
+	  TYPE_HAS_CONSTRUCTOR (targ) = 0;
 	  return build_up_reference (type, temp, flags);
 	}
       /* Let &* cancel out to simplify resulting code.
          Also, throw away intervening NOP_EXPRs.  */
-      arg = TREE_OPERAND (arg, 0);
+      arg = TREE_OPERAND (targ, 0);
       if (TREE_CODE (arg) == NOP_EXPR || TREE_CODE (arg) == REFERENCE_EXPR)
 	arg = TREE_OPERAND (arg, 0);
 
@@ -254,20 +257,20 @@ build_up_reference (type, arg, flags)
 	 Also, build up references to non-lvalues it we must.  */
       /* For &x[y], return (&) x+y */
     case ARRAY_REF:
-      if (mark_addressable (TREE_OPERAND (arg, 0)) == 0)
+      if (mark_addressable (TREE_OPERAND (targ, 0)) == 0)
 	return error_mark_node;
-      rval = build_binary_op (PLUS_EXPR, TREE_OPERAND (arg, 0),
-			      TREE_OPERAND (arg, 1));
+      rval = build_binary_op (PLUS_EXPR, TREE_OPERAND (targ, 0),
+			      TREE_OPERAND (targ, 1));
       TREE_TYPE (rval) = type;
-      if (TREE_LITERAL (TREE_OPERAND (arg, 1))
-	  && staticp (TREE_OPERAND (arg, 0)))
+      if (TREE_LITERAL (TREE_OPERAND (targ, 1))
+	  && staticp (TREE_OPERAND (targ, 0)))
 	TREE_LITERAL (rval) = 1;
       return rval;
 
     case SCOPE_REF:
       /* Could be a reference to a static member.  */
       {
-	tree field = TREE_OPERAND (arg, 1);
+	tree field = TREE_OPERAND (targ, 1);
 	if (TREE_STATIC (field))
 	  {
 	    rval = build1 (ADDR_EXPR, type, field);
@@ -279,10 +282,10 @@ build_up_reference (type, arg, flags)
       assert (0);
 
     case COMPONENT_REF:
-      rval = build_component_addr (arg, build_pointer_type (argtype),
+      rval = build_component_addr (targ, build_pointer_type (argtype),
 				   "attempt to make a reference to bit-field structure member `%s'");
-      rval = build1 (NOP_EXPR, type, rval);
-      literal_flag = staticp (TREE_OPERAND (arg, 0));
+      TREE_TYPE (rval) = type;
+      literal_flag = staticp (TREE_OPERAND (targ, 0));
 #if 0
       goto done_but_maybe_warn;
 #else
@@ -296,73 +299,74 @@ build_up_reference (type, arg, flags)
 	 as `register'.  */
 
     case RESULT_DECL:
-      if (staticp (arg))
+      if (staticp (targ))
 	literal_flag = 1;
-      TREE_ADDRESSABLE (arg) = 1;
-      put_var_into_stack (arg);
+      TREE_ADDRESSABLE (targ) = 1;
+      put_var_into_stack (targ);
       break;
 
     case PARM_DECL:
-      if (arg == current_class_decl)
+      if (targ == current_class_decl)
 	{
 	  error ("address of `this' not available");
-	  TREE_ADDRESSABLE (arg) = 1; /* so compiler doesn't die later */
-	  put_var_into_stack (arg);
+	  TREE_ADDRESSABLE (targ) = 1; /* so compiler doesn't die later */
+	  put_var_into_stack (targ);
 	  break;
 	}
       /* Fall through.  */
     case VAR_DECL:
     case CONST_DECL:
-      if (TREE_REGDECL (arg) && !TREE_ADDRESSABLE (arg))
+      if (TREE_REGDECL (targ) && !TREE_ADDRESSABLE (targ))
 	warning ("address needed to build reference for `%s', which is declared `register'",
-		 IDENTIFIER_POINTER (DECL_NAME (arg)));
-      else if (staticp (arg))
+		 IDENTIFIER_POINTER (DECL_NAME (targ)));
+      else if (staticp (targ))
 	literal_flag = 1;
 
-      TREE_ADDRESSABLE (arg) = 1;
-      put_var_into_stack (arg);
+      TREE_ADDRESSABLE (targ) = 1;
+      put_var_into_stack (targ);
       break;
 
     case COMPOUND_EXPR:
       {
-	tree real_reference = build_up_reference (type, TREE_OPERAND (arg, 1), 1);
-	rval = build (COMPOUND_EXPR, type, TREE_OPERAND (arg, 0), real_reference);
-	TREE_LITERAL (rval) = staticp (TREE_OPERAND (arg, 1));
+	tree real_reference = build_up_reference (type, TREE_OPERAND (targ, 1),
+						  LOOKUP_PROTECT);
+	rval = build (COMPOUND_EXPR, type, TREE_OPERAND (targ, 0), real_reference);
+	TREE_LITERAL (rval) = staticp (TREE_OPERAND (targ, 1));
 	return rval;
       }
 
     case MODIFY_EXPR:
     case INIT_EXPR:
       {
-	tree real_reference = build_up_reference (type, TREE_OPERAND (arg, 0), 1);
+	tree real_reference = build_up_reference (type, TREE_OPERAND (targ, 0),
+						  LOOKUP_PROTECT);
 	rval = build (COMPOUND_EXPR, type, arg, real_reference);
-	TREE_LITERAL (rval) = staticp (TREE_OPERAND (arg, 0));
+	TREE_LITERAL (rval) = staticp (TREE_OPERAND (targ, 0));
 	return rval;
       }
 
     case COND_EXPR:
       return build (COND_EXPR, type,
-		    TREE_OPERAND (arg, 0),
-		    build_up_reference (type, TREE_OPERAND (arg, 1), 1),
-		    build_up_reference (type, TREE_OPERAND (arg, 2), 1));
+		    TREE_OPERAND (targ, 0),
+		    build_up_reference (type, TREE_OPERAND (targ, 1), LOOKUP_PROTECT),
+		    build_up_reference (type, TREE_OPERAND (targ, 2), LOOKUP_PROTECT));
 
     case WITH_CLEANUP_EXPR:
-      rval = build (WITH_CLEANUP_EXPR, type,
-		    build_up_reference (type, TREE_OPERAND (arg, 0), 1),
-		    0, TREE_OPERAND (arg, 2));
-      return rval;
+      return build (WITH_CLEANUP_EXPR, type,
+		    build_up_reference (type, TREE_OPERAND (targ, 0), LOOKUP_PROTECT),
+		    0, TREE_OPERAND (targ, 2));
 
     default:
       break;
     }
 
-  if (TREE_ADDRESSABLE (arg) == 0)
+  if (TREE_ADDRESSABLE (targ) == 0)
     {
       tree temp;
 
-      if (TREE_CODE (arg) == CALL_EXPR && IS_AGGR_TYPE (argtype))
+      if (TREE_CODE (targ) == CALL_EXPR && IS_AGGR_TYPE (argtype))
 	{
-	  temp = build_cplus_new (argtype, arg);
+	  temp = build_cplus_new (argtype, targ, 1);
 	  rval = build1 (ADDR_EXPR, type, temp);
 	  goto done;
 	}
@@ -372,9 +376,9 @@ build_up_reference (type, arg, flags)
 	  if (global_bindings_p ())
 	    {
 	      /* Give this new temp some rtl and initialize it.  */
-	      DECL_INITIAL (temp) = arg;
+	      DECL_INITIAL (temp) = targ;
 	      TREE_STATIC (temp) = 1;
-	      finish_decl (temp, arg, NULL_TREE);
+	      finish_decl (temp, targ, NULL_TREE);
 	      /* Do this after declaring it static.  */
 	      rval = build_unary_op (ADDR_EXPR, temp, 0);
 	      literal_flag = TREE_LITERAL (rval);
@@ -389,7 +393,7 @@ build_up_reference (type, arg, flags)
 		  /* This may produce surprising results,
 		     since we commit to initializing the temp
 		     when the temp may not actually get used.  */
-		  expand_aggr_init (temp, arg, 0);
+		  expand_aggr_init (temp, targ, 0);
 		  TREE_TYPE (rval) = type;
 		  literal_flag = TREE_LITERAL (rval);
 		  goto done;
@@ -407,7 +411,11 @@ build_up_reference (type, arg, flags)
 	}
     }
   else
-    rval = build1 (ADDR_EXPR, type, arg);
+    {
+      if (TREE_CODE (arg) == SAVE_EXPR)
+	abort ();
+      rval = build1 (ADDR_EXPR, type, arg);
+    }
 
  done_but_maybe_warn:
   if (TREE_READONLY (arg)
@@ -421,7 +429,7 @@ build_up_reference (type, arg, flags)
     {
       TREE_TYPE (rval) = TYPE_POINTER_TO (argtype);
       rval = convert_pointer_to (target_type, rval);
-      rval = build1 (NOP_EXPR, type, rval);
+      TREE_TYPE (rval) = type;
     }
   TREE_LITERAL (rval) = literal_flag;
   return rval;
@@ -464,7 +472,10 @@ convert_to_reference (decl, reftype, expr, strict, flags)
       else if (rval = build_type_conversion (CONVERT_EXPR, type, expr, 1))
 	{
 	  if (TYPE_NEEDS_DESTRUCTOR (type))
-	    rval = cleanup_after_call (rval);
+	    {
+	      rval = cleanup_after_call (rval);
+	      rval = convert_to_reference (NULL_TREE, reftype, rval, strict, flags);
+	    }
 	  else
 	    {
 	      decl = get_temp_name (type, 0);
@@ -472,7 +483,6 @@ convert_to_reference (decl, reftype, expr, strict, flags)
 	      rval = build (COMPOUND_EXPR, reftype, rval,
 			    convert_to_reference (NULL_TREE, reftype, decl,
 						  strict, flags));
-
 	    }
 	  return rval;
 	}
@@ -548,7 +558,7 @@ convert_to_reference (decl, reftype, expr, strict, flags)
 
       if (init == error_mark_node)
 	return error_mark_node;
-      rval = build_cplus_new (type, init);
+      rval = build_cplus_new (type, init, 1);
       if (decl == error_mark_node)
 	flags |= LOOKUP_PROTECTED_OK;
       return build_up_reference (reftype, rval, flags);
@@ -569,32 +579,24 @@ convert_from_reference (val)
 {
   tree type = TREE_TYPE (val);
 
-#if 0
-  if (TREE_CODE (val) == REFERENCE_EXPR)
-    {
-      val = build1 (NOP_EXPR, build_pointer_type (TREE_TYPE (type)),
-		    TREE_OPERAND (val, 0));
-      return val;
-    }
-#endif
   if (TREE_CODE (type) == OFFSET_TYPE)
     type = TREE_TYPE (type);
  if (TREE_CODE (type) == REFERENCE_TYPE)
     {
-      tree dt = TREE_TYPE (type);
+      tree target_type = TREE_TYPE (type);
 
       /* This can happen if we cast to a reference type.  */
       if (TREE_CODE (val) == ADDR_EXPR)
 	{
-	  val = build1 (NOP_EXPR, build_pointer_type (dt), val);
+	  val = build1 (NOP_EXPR, build_pointer_type (target_type), val);
 	  val = build_indirect_ref (val, 0);
 	  return val;
 	}
 
-      val = build1 (INDIRECT_REF, TYPE_MAIN_VARIANT (dt), val);
+      val = build1 (INDIRECT_REF, TYPE_MAIN_VARIANT (target_type), val);
 
-      TREE_THIS_VOLATILE (val) = TREE_VOLATILE (dt);
-      TREE_READONLY (val) = TREE_READONLY (dt);
+      TREE_THIS_VOLATILE (val) = TREE_VOLATILE (target_type);
+      TREE_READONLY (val) = TREE_READONLY (target_type);
     }
   return val;
 }
@@ -861,10 +863,10 @@ convert_to_integer (type, expr)
 		    typex = (TREE_UNSIGNED (TREE_TYPE (expr))
 			     ? unsigned_type (typex) : signed_type (typex));
 		    return convert (type,
-				    build (COND_EXPR, typex,
-					   TREE_OPERAND (expr, 0),
-					   convert (typex, arg1),
-					   convert (typex, arg2)));
+				    fold (build (COND_EXPR, typex,
+						 TREE_OPERAND (expr, 0),
+						 convert (typex, arg1),
+						 convert (typex, arg2))));
 		  }
 	      }
 	  }
@@ -914,8 +916,7 @@ convert_to_aggr (type, expr, msgp, protect)
 {
   tree basetype = TYPE_MAIN_VARIANT (type);
   tree name = DECL_NAME (TYPE_NAME (basetype));
-  tree field;
-  tree function, fntype, parmtypes, parmlist, result;
+  tree function, fndecl, fntype, parmtypes, parmlist, result;
   tree method_name;
   enum visibility_type visibility;
   int can_be_private, can_be_protected;
@@ -947,27 +948,27 @@ convert_to_aggr (type, expr, msgp, protect)
   method_name = build_decl_overload (IDENTIFIER_POINTER (name), parmtypes, 1);
 
   /* constructors are up front.  */
-  field = TREE_VEC_ELT (CLASSTYPE_METHOD_VEC (basetype), 0);
+  fndecl = TREE_VEC_ELT (CLASSTYPE_METHOD_VEC (basetype), 0);
   if (TYPE_HAS_DESTRUCTOR (basetype))
-    field = TREE_CHAIN (field);
+    fndecl = TREE_CHAIN (fndecl);
 
-  while (field)
+  while (fndecl)
     {
-      if (DECL_NAME (field) == method_name)
+      if (DECL_NAME (fndecl) == method_name)
 	{
-	  function = field;
+	  function = fndecl;
 	  if (protect)
 	    {
-	      if (TREE_PRIVATE (field))
+	      if (TREE_PRIVATE (fndecl))
 		{
 		  can_be_private =
 		    (basetype == current_class_type
 		     || is_friend (basetype, current_function_decl)
-		     || purpose_member (basetype, DECL_VISIBILITY (field)));
+		     || purpose_member (basetype, DECL_VISIBILITY (fndecl)));
 		  if (! can_be_private)
 		    goto found;
 		}
-	      else if (TREE_PROTECTED (field))
+	      else if (TREE_PROTECTED (fndecl))
 		{
 		  if (! can_be_protected)
 		    goto found;
@@ -975,35 +976,35 @@ convert_to_aggr (type, expr, msgp, protect)
 	    }
 	  goto found_and_ok;
 	}
-      field = TREE_CHAIN (field);
+      fndecl = TREE_CHAIN (fndecl);
     }
 
   /* No exact conversion was found.  See if an approximate
      one will do.  */
-  field = TREE_VEC_ELT (CLASSTYPE_METHOD_VEC (basetype), 0);
+  fndecl = TREE_VEC_ELT (CLASSTYPE_METHOD_VEC (basetype), 0);
   if (TYPE_HAS_DESTRUCTOR (basetype))
-    field = TREE_CHAIN (field);
+    fndecl = TREE_CHAIN (fndecl);
 
   {
     int saw_private = 0;
     int saw_protected = 0;
     struct candidate *candidates =
-      (struct candidate *) alloca ((list_length (field)+1) * sizeof (struct candidate));
+      (struct candidate *) alloca ((list_length (fndecl)+1) * sizeof (struct candidate));
     struct candidate *cp = candidates;
 
-    while (field)
+    while (fndecl)
       {
-	function = field;
+	function = fndecl;
 	cp->harshness = (unsigned short *)alloca (3 * sizeof (short));
-	compute_conversion_costs (function, parmlist, cp, 2);
+	compute_conversion_costs (fndecl, parmlist, cp, 2);
 	if (cp->evil == 0)
 	  {
-	    cp->u.field = field;
+	    cp->u.field = fndecl;
 	    if (protect)
 	      {
-		if (TREE_PRIVATE (field))
+		if (TREE_PRIVATE (fndecl))
 		  visibility = visibility_private;
-		else if (TREE_PROTECTED (field))
+		else if (TREE_PROTECTED (fndecl))
 		  visibility = visibility_protected;
 		else
 		  visibility = visibility_public;
@@ -1014,10 +1015,10 @@ convert_to_aggr (type, expr, msgp, protect)
 	    if (visibility == visibility_private
 		? (basetype == current_class_type
 		   || is_friend (basetype, cp->function)
-		   || purpose_member (basetype, DECL_VISIBILITY (field)))
+		   || purpose_member (basetype, DECL_VISIBILITY (fndecl)))
 		: visibility == visibility_protected
 		? (can_be_protected
-		   || purpose_member (basetype, DECL_VISIBILITY (field)))
+		   || purpose_member (basetype, DECL_VISIBILITY (fndecl)))
 		: 1)
 	      {
 		if (cp->user == 0 && cp->b_or_d == 0
@@ -1035,7 +1036,7 @@ convert_to_aggr (type, expr, msgp, protect)
 		  saw_protected = 1;
 	      }
 	  }
-	field = TREE_CHAIN (field);
+	fndecl = TREE_CHAIN (fndecl);
       }
     if (cp - candidates)
       {
@@ -1059,7 +1060,7 @@ convert_to_aggr (type, expr, msgp, protect)
 	  }
 
 	function = cp->function;
-	field = cp->u.field;
+	fndecl = cp->u.field;
 	goto found_and_ok;
       }
     else if (msgp)
@@ -1084,7 +1085,7 @@ convert_to_aggr (type, expr, msgp, protect)
     if (! can_be_private)
       {
 	if (msgp)
-	  *msgp = TREE_PRIVATE (field)
+	  *msgp = TREE_PRIVATE (fndecl)
 	    ? "conversion to type `%s' is private"
 	    : "conversion to type `%s' is from private base class";
 	return error_mark_node;
@@ -1093,12 +1094,12 @@ convert_to_aggr (type, expr, msgp, protect)
     if (! can_be_protected)
       {
 	if (msgp)
-	  *msgp = TREE_PRIVATE (field)
+	  *msgp = TREE_PRIVATE (fndecl)
 	    ? "conversion to type `%s' is protected"
 	    : "conversion to type `%s' is from protected base class";
 	return error_mark_node;
       }
-  function = field;
+  function = fndecl;
  found_and_ok:
 
   /* It will convert, but we don't do anything about it yet.  */
@@ -1282,7 +1283,7 @@ convert (type, expr)
 		      return error_mark_node;
 		    }
 		  /* call to constructor successful.  */
-		  rval = build_cplus_new (type, rval);
+		  rval = build_cplus_new (type, rval, 0);
 		  return rval;
 		}
 	    }
@@ -1321,7 +1322,7 @@ convert (type, expr)
 	      error_with_aggr_type (type, "in conversion to type `%s'");
 	      return error_mark_node;
 	    }
-	  rval = build_cplus_new (type, init);
+	  rval = build_cplus_new (type, init, 0);
 	  return rval;
 	}
     }
@@ -1381,7 +1382,8 @@ build_type_conversion_1 (xtype, basetype, expr, typename, for_sure)
     }
   if (first_arg != expr)
     {
-      expr = build_up_reference (build_reference_type (TREE_TYPE (expr)), expr, 0);
+      expr = build_up_reference (build_reference_type (TREE_TYPE (expr)), expr,
+				 LOOKUP_COMPLAIN);
       TREE_VALUE (TREE_OPERAND (rval, 1)) = build_unary_op (ADDR_EXPR, expr, 0);
     }
   if (TREE_CODE (TREE_TYPE (rval)) == REFERENCE_TYPE
@@ -1504,8 +1506,9 @@ build_type_conversion (code, xtype, expr, for_sure)
 	      if (IS_AGGR_TYPE (type))
 		{
 		  tree init = build_method_call (NULL_TREE, DECL_NAME (TYPE_NAME (type)), build_tree_list (NULL_TREE, rval), NULL_TREE, LOOKUP_NORMAL);
-		  tree temp = build_cplus_new (type, init);
-		  return build_up_reference (TYPE_REFERENCE_TO (type), temp, 0);
+		  tree temp = build_cplus_new (type, init, 1);
+		  return build_up_reference (TYPE_REFERENCE_TO (type), temp,
+					     LOOKUP_COMPLAIN);
 		}
 	      return convert (xtype, rval);
 	    }
@@ -1743,16 +1746,16 @@ build_default_binary_type_conversion (code, arg1, arg2)
   name1 = TYPE_NAME_STRING (type1);
   name2 = TYPE_NAME_STRING (type2);
 
-  if (! TYPE_HAS_CONVERSION (type1))
+  if (!IS_AGGR_TYPE (type1) || !TYPE_HAS_CONVERSION (type1))
     {
-      if (! TYPE_HAS_CONVERSION (type2))
+      if (!IS_AGGR_TYPE (type2) || !TYPE_HAS_CONVERSION (type2))
 	error ("type conversion required for binary operation on types `%s' and `%s'",
 	       name1, name2);
       else
 	error ("type conversion required for type `%s'", name1);
       return 0;
     }
-  else if (! TYPE_HAS_CONVERSION (type2))
+  else if (!IS_AGGR_TYPE (type2) || !TYPE_HAS_CONVERSION (type2))
     {
       error ("type conversion required for type `%s'", name2);
       return 0;
