@@ -46,13 +46,14 @@
 /* #define nounderscore 1 /* define this is names don't start with _ */
 
 #include <stdio.h>
-#include <string.h>
 #include <ctype.h>
 
 #ifdef USG
 #include <memory.h>
+#include <string.h>
 #else
-#define memcpy(s1, s2, n) bcopy ((s2), (s1), (n)) 
+#include <strings.h>
+#define memcpy(s1, s2, n) bcopy ((s2), (s1), (n))
 #define memcmp(s1, s2, n) bcmp ((s2), (s1), (n))
 #define strchr index 
 #define strrchr rindex
@@ -148,6 +149,7 @@ static int do_type (const char **type, string *result);
 static int do_arg (const char **type, string *result);
 static int do_args (const char **type, string *decl);
 static void munge_function_name (string *name);
+static void remember_type (const char *type, int len);
 #else
 static void string_need ();
 static void string_delete ();
@@ -166,6 +168,7 @@ static int do_type ();
 static int do_arg ();
 static int do_args ();
 static void munge_function_name ();
+static void remember_type ();
 #endif
 
 char *
@@ -178,7 +181,10 @@ cplus_demangle (type)
   int constructor = 0;
   int const_flag = 0;
   int i;
-  const char *p, *premangle;
+  const char *p;
+#ifndef LONGERNAMES
+  const char *premangle;
+#endif
 
   if (type == NULL || *type == '\0')
     return NULL;
@@ -242,7 +248,9 @@ cplus_demangle (type)
     }
   p += 2;
 
+#ifndef LONGERNAMES
   premangle = p;
+#endif
   switch (*p)
     {
     case 'C':
@@ -289,10 +297,9 @@ cplus_demangle (type)
 	  string_prepend (&decl, "::");
 	  string_prependn (&decl, p, n);
 	}
-#ifndef LONGERNAMES
-      p = premangle;
-#else
       p += n;
+#ifndef LONGERNAMES
+      remember_type (premangle, p - premangle);
 #endif
       success = do_args (&p, &decl);
       if (const_flag)
@@ -554,6 +561,12 @@ do_type (type, result)
 	  string_append (result, " ");
 	string_append (result, "void");
 	break;
+      case 'x':
+	*type += 1;
+	if (non_empty)
+	  string_append (result, " ");
+	string_append (result, "long long");
+	break;
       case 'l':
 	*type += 1;
 	if (non_empty)
@@ -662,15 +675,21 @@ do_arg (type, result)
      const char **type;
      string *result;
 {
-  char *tem;
-  int len;
-  const char *start;
-  const char *end;
+  const char *start = *type;
 
-  start = *type;
   if (!do_type (type, result))
     return 0;
-  end = *type;
+  remember_type (start, *type - start);
+  return 1;
+}
+
+static void
+remember_type (start, len)
+     const char *start;
+     int len;
+{
+  char *tem;
+
   if (ntypes >= typevec_size)
     {
       if (typevec_size == 0)
@@ -684,12 +703,10 @@ do_arg (type, result)
 	  typevec = (char **) xrealloc ((char *)typevec, sizeof (char*)*typevec_size);
 	}
     }
-  len = end - start;
   tem = (char *) xmalloc (len + 1);
   memcpy (tem, start, len);
   tem[len] = '\0';
   typevec[ntypes++] = tem;
-  return 1;
 }
 
 /* `decl' must be already initialised, usually non-empty;
@@ -702,13 +719,6 @@ do_args (type, decl)
 {
   string arg;
   int need_comma = 0;
-  int dont_want_first;
-
-#ifndef LONGERNAMES
-  dont_want_first = 1;
-#else
-  dont_want_first = 0;
-#endif
 
   string_append (decl, "(");
 
@@ -739,14 +749,9 @@ do_args (type, decl)
 	    string_append (decl, ", ");
 	  if (!do_arg (type, &arg))
 	    return 0;
-	  if (dont_want_first)
-	    dont_want_first = 0;
-	  else
-	    {
-	      string_appends (decl, &arg);
-	      need_comma = 1;
-	    }
+	  string_appends (decl, &arg);
 	  string_delete (&arg);
+	  need_comma = 1;
 	}
     }
 
