@@ -22,18 +22,8 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "tree.h"
 #include <stdio.h>
 
+extern char **tree_code_name;
 
-/* Names of tree components.
-   Used for printing out the tree and error messages.  */
-#define DEFTREECODE(SYM, NAME, TYPE, LEN) NAME,
-
-char *tree_code_name[] = {
-#include "tree.def"
-};
-#undef DEFTREECODE
-
-extern char *tree_code_type[];
-extern int tree_code_length[];
 extern char *mode_name[];
 
 extern char spaces[];
@@ -58,6 +48,14 @@ debug_dump_tree (root)
   dump_tree (stderr, root);
 }
 
+/* Don't know, why they have different names in C and c++ (hgs) */
+void
+debug_tree (root)
+     tree root;
+{
+  dump_tree (stderr, root);
+}
+
 void
 dump_tree (outf, root)
      FILE *outf;
@@ -246,6 +244,36 @@ prtypeinfo (node)
       fputs ("used", outfile);
       first = 0;
     }
+  if (TREE_PERMANENT (node))
+    {
+      if (!first) putc (' ', outfile);
+      fputs ("permanent", outfile);
+      first = 0;
+    }
+  if (TREE_LANG_FLAG_1 (node))
+    {
+      if (!first) putc (' ', outfile);
+      fputs ("lang_flag_1", outfile);
+      first = 0;
+    }
+  if (TREE_LANG_FLAG_2 (node))
+    {
+      if (!first) putc (' ', outfile);
+      fputs ("lang_flag_2", outfile);
+      first = 0;
+    }
+  if (TREE_LANG_FLAG_3 (node))
+    {
+      if (!first) putc (' ', outfile);
+      fputs ("lang_flag_3", outfile);
+      first = 0;
+    }
+  if (TREE_LANG_FLAG_4 (node))
+    {
+      if (!first) putc (' ', outfile);
+      fputs ("lang_flag_4", outfile);
+      first = 0;
+    }
   fputs ("] ", outfile);
 }
 
@@ -463,9 +491,12 @@ dump (node, indent)
 	{
 	  if (i >= first_rtl)
 	    {
-	      fprintf (outfile, "\n");
 	      skip (indent);
-	      print_rtl (outfile, TREE_OPERAND (node, i));
+	      if (TREE_OPERAND (node, i))
+		print_rtl (outfile, TREE_OPERAND (node, i));
+	      else
+		fprintf (outfile, "(nil)");
+	      fprintf (outfile, "\n");
 	    }
 	  else
 	    {
@@ -477,7 +508,7 @@ dump (node, indent)
       part ("chain", TREE_CHAIN (node));
       fputc ('\n', outfile);
       walk (TREE_TYPE (node), node, indent);
-      for (i = 0; i < len; i++)
+      for (i = 0; i < len && i < first_rtl; i++)
 	walk (TREE_OPERAND (node, i), node, indent);
       break;
 
@@ -500,6 +531,7 @@ dump (node, indent)
 	  part ("supercontext", STMT_SUPERCONTEXT (node));
 	  part ("bind_size", STMT_BIND_SIZE (node));
 	  part ("body", STMT_BODY (node));
+	  part ("subblocks", STMT_SUBBLOCKS (node));
 	  break;
 
 	case CASE_STMT:
@@ -529,6 +561,7 @@ dump (node, indent)
 	  walk (STMT_SUPERCONTEXT (node), node, indent);
 	  walk (STMT_BIND_SIZE (node), node, indent);
 	  walk (STMT_BODY (node), node, indent);
+	  walk (STMT_SUBBLOCKS (node), node, indent);
 	  break;
 
 	case CASE_STMT:
@@ -563,8 +596,8 @@ dump (node, indent)
 #else
 	  {
 	    int i;
-	    fprintf (outfile, " = 0x");
 	    char *p = (char *) &TREE_REAL_CST (node);
+	    fprintf (outfile, " = 0x");
 	    for (i = 0; i < sizeof TREE_REAL_CST (node); i++)
 	      fprintf (outfile, "%02x", *p++);
 	    fprintf (outfile, ";");
